/**
 * Represents a wrapped constructor of a model, which is essentially
 * just the constructor with some merged definitions that add static methods.
 */
import QueryBuilder from "./query/builder";
export declare type Wrapped<T, A> = T & {
    readonly tableName: string;
} & QueryBuilder<A>;
export declare type R<T> = QueryBuilder<T> & (() => Promise<T>);
export declare type A<T> = QueryBuilder<T> & (() => Promise<T[]>);
/**
 * Represents the types currently supported natively.
 */
export declare type DatabaseType = number | string | boolean;
export declare type KeyedDatabaseResult = {
    [key: string]: DatabaseType;
};
/**
 * Represents an abstract base model that every model inherits from.
 */
export default abstract class BaseModel {
    readonly id: number;
    /**
     * Returns a new builder that contains a WHERE clause for the current id.
     */
    builder(): QueryBuilder<this>;
    /**
     * Saves this object, inserting it if it doesn't exist or updating it otherwise.
     */
    save(): Promise<void>;
    /**
     * Deletes this object, throwing if it is not currently stored.
     */
    delete(): Promise<void>;
    /**
     * Indicates that this object has a single associated object of the provided
     * model type and optional foreign key (which will default to the name of the
     * current model with _id).
     */
    hasOne<A, T extends Wrapped<any, A>>(model: T, foreignKey?: string): R<A>;
    /**
     * Indicates that this object has a single associated object of the provided
     * model type. The object will be looked up by looking at the value of the specified
     * column and finding the row of the specified model with that ID.
     */
    belongsTo<A, T extends Wrapped<any, A>, K extends keyof this>(model: T, foreignKey?: K): R<A>;
    /**
     * Relationship that finds all members of the specified model table with their
     * foreign key set to the ID of the current object.
     */
    hasMany<Q, T extends Wrapped<any, Q>>(model: T, foreignKey?: string): A<Q>;
    /**
     * Relationship that functions as a hasMany going both ways. This requires a pivot table
     * which joins the two models together.
     */
    hasAndBelongsToMany<Q, T extends Wrapped<any, Q>>(model: T, pivotName?: string): A<Q>;
}
