"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
/**
 * Represents a wrapped constructor of a model, which is essentially
 * just the constructor with some merged definitions that add static methods.
 */
const builder_1 = require("./query/builder");
const util_1 = require("./util");
/**
 * Represents an abstract base model that every model inherits from.
 */
class BaseModel {
    /**
     * Returns a new builder that contains a WHERE clause for the current id.
     */
    builder() {
        return builder_1.default.table(Object.getPrototypeOf(this).tableName).where("id", this.id);
    }
    /**
     * Saves this object, inserting it if it doesn't exist or updating it otherwise.
     */
    save() {
        // Insert
        if (typeof this.id === "undefined") {
            return this.builder().insertAndGetId(this).then(id => {
                this.id = id;
            });
        }
        else {
            return this.builder().update(this);
        }
    }
    /**
     * Deletes this object, throwing if it is not currently stored.
     */
    delete() {
        if (typeof this.id === "undefined")
            throw new Error("Cannot delete object if it is not in the database.");
        return this.builder().delete().then(x => {
            this.id = undefined;
        });
    }
    /**
     * Indicates that this object has a single associated object of the provided
     * model type and optional foreign key (which will default to the name of the
     * current model with _id).
     */
    hasOne(model, foreignKey) {
        let instance;
        return new Proxy(/* istanbul ignore next */ () => { }, {
            apply: () => {
                if (instance)
                    return instance;
                return instance = model.where(foreignKey || util_1.getForeignKey(this.constructor.name), this.id).limit(1).first();
            },
            get: (obj, key) => {
                const builder = model.where(foreignKey || util_1.getForeignKey(this.constructor.name), this.id).limit(1);
                if (typeof builder[key] === "undefined")
                    return undefined;
                return typeof builder[key] === "function" ? builder[key].bind(builder) : builder[key];
            }
        });
    }
    /**
     * Indicates that this object has a single associated object of the provided
     * model type. The object will be looked up by looking at the value of the specified
     * column and finding the row of the specified model with that ID.
     */
    belongsTo(model, foreignKey) {
        let instance;
        return new Proxy(/* istanbul ignore next */ () => { }, {
            apply: () => {
                if (instance)
                    return instance;
                return instance = model.where("id", this[foreignKey || util_1.getForeignKey(model.name)]).first();
            },
            get: (obj, key) => {
                const builder = model.where("id", this[foreignKey || util_1.getForeignKey(model.name)]);
                if (typeof builder[key] === "undefined")
                    return undefined;
                return typeof builder[key] === "function" ? builder[key].bind(builder) : builder[key];
            }
        });
    }
    /**
     * Relationship that finds all members of the specified model table with their
     * foreign key set to the ID of the current object.
     */
    hasMany(model, foreignKey) {
        let instance;
        return new Proxy(/* istanbul ignore next */ () => { }, {
            apply: () => {
                if (instance)
                    return instance;
                return instance = model.where(foreignKey || util_1.getForeignKey(this.constructor.name), this.id).all();
            },
            get: (obj, key) => {
                const builder = model.where(foreignKey || util_1.getForeignKey(this.constructor.name), this.id);
                if (typeof builder[key] === "undefined")
                    return undefined;
                return typeof builder[key] === "function" ? builder[key].bind(builder) : builder[key];
            }
        });
    }
    /**
     * Relationship that functions as a hasMany going both ways. This requires a pivot table
     * which joins the two models together.
     */
    hasAndBelongsToMany(model, pivotName) {
        let instance;
        const pivot = pivotName || util_1.getPivotTableName(this.constructor.name, model.name);
        const createBuilder = () => {
            const b = model
                .join(pivot, `${model.tableName}.id`, "=", `${pivot}.${util_1.getForeignKey(model.name)}`)
                .where(`${pivot}.${util_1.getForeignKey(this.constructor.name)}`, this.id);
            b.modelConstructor = model;
            return b;
        };
        return new Proxy(/* istanbul ignore next */ () => { }, {
            apply: () => {
                if (instance)
                    return instance;
                return instance = createBuilder().all();
            },
            get: (obj, key) => {
                const builder = createBuilder();
                if (typeof builder[key] === "undefined")
                    return undefined;
                return typeof builder[key] === "function" ? builder[key].bind(builder) : builder[key];
            }
        });
    }
}
exports.default = BaseModel;
