"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const sqlite3_engine_1 = require("./engines/sqlite3-engine");
const postgres_engine_1 = require("./engines/postgres-engine");
const util_1 = require("./util");
const builder_1 = require("./query/builder");
const mysql_engine_1 = require("./engines/mysql-engine");
class BasieStatic {
    constructor() {
        this.engine = null;
    }
    /**
     * Returns the engine currently in use, or throws if there is no engine configured.
     */
    getEngine() {
        if (!this.engine)
            throw new Error("No database engine has been configured. Ensure that you use Basie.use(engine) before attempting any queries.");
        return this.engine;
    }
    /**
     * Changes the current database engine to the specified engine.
     */
    use(engine) {
        this.engine = engine;
    }
    /**
     * Configures basie to be used with the specified sqlite database.
     */
    sqlite(db) {
        this.use(new sqlite3_engine_1.default(db));
    }
    /**
     * Configures basie to be used with the specified postgres connection pool.
     */
    postgres(db) {
        this.use(new postgres_engine_1.default(db));
    }
    /**
     * Configures basie to be used with the specified mysql connection.
     */
    mysql(connection) {
        this.use(new mysql_engine_1.default(connection));
    }
    /**
     * Wraps the specified model with basie-specific methods that give it static
     * querying methods. This does not create a table within the database, and
     * instead expects the table to already be present. This function is called
     * a bit weirdly (`Basie.wrap<_Model>()(_Model)`) to overcome some typescript
     * type system limits and statically typecheck that you're only using valid
     * members.
     */
    wrap() {
        return (arg, tableName) => {
            if (!tableName)
                tableName = util_1.getTableName(arg.name);
            const local = arg;
            local.tableName = tableName;
            local.prototype.tableName = tableName;
            return new Proxy(local, {
                get(target, name) {
                    // If the property exists on the object itself (static method), return that.
                    /* istanbul ignore next This incorrectly reports as uncovered, even though it runs. */
                    if (typeof target[name] !== "undefined")
                        return target[name];
                    const builder = builder_1.default.model(local);
                    // If it is not a function, act like it doesn't exist (which is probably the thruth).
                    /* istanbul ignore next This incorrectly reports as uncovered, even though it runs. */
                    if (typeof builder[name] !== "function")
                        return undefined;
                    // Return that method of the QueryBuilder and bind it to the builder.
                    return builder[name].bind(builder);
                }
            });
        };
    }
}
exports.BasieStatic = BasieStatic;
exports.default = new BasieStatic();
var base_model_1 = require("./base-model");
exports.BaseModel = base_model_1.default;
