import { JoinType, Operator, OrderDirection, QueryBoolean, WhereQueryClause } from "./types";
import { DatabaseType, KeyedDatabaseResult, Wrapped } from "../base-model";
/**
 * A fluid SQL query builder that is semi-typesafe and supports selecting, inserting, updating
 * and deleting. Obtain an instance using the static methods or one of the methods on wrapped
 * models.
 */
export default class QueryBuilder<T> {
    /**
     * The constructor of the model we are currently querying. If this is non-null,
     * we are still operating on the full model (no narrowed down views) which means
     * that once we have fetched the objects we can assign this prototype to ensure that
     * the returned instance supports any user-defined methods (along with defaults such as save).
     */
    modelConstructor: Function | null;
    /**
     * The aggregate function for this query. If this is set, it is assumed
     * that this query selects of format `SELECT fun(columns) AS aggregate`.
     */
    aggregateFunction: string | null;
    /**
     * Which table we are currently querying.
     */
    table: string;
    /**
     * The columns we are currently retrieving.
     */
    columns: string[];
    /**
     * If we are only looking for distinct values.
     */
    isDistinct: boolean;
    /**
     * All joins for the current query.
     */
    joins: JoinClause[];
    /**
     * All where clauses for the current query.
     */
    wheres: WhereQueryClause[];
    /**
     * All the GROUP BY columns for the query.
     */
    groups: string[];
    /**
     * All the ORDER BY clauses for the query.
     */
    orders: {
        column: string;
        direction: OrderDirection;
    }[];
    /**
     * The LIMIT clause for the query, or -1 if not applicable.
     */
    limitCount: number;
    protected constructor();
    /**
     * Narrows this query down to the specified fields. This erases type-
     * safety since raw SQL queries can be used. Do not use with untrusted input.
     */
    select(...fields: string[]): QueryBuilder<KeyedDatabaseResult>;
    /**
     * Marks this query as looking for distinct values. Use distinct(false) to undo.
     */
    distinct(value?: boolean): this;
    /**
     * Changes the table this query is currently acting on to the table associated
     * with the specified model.
     */
    from<A, M extends Wrapped<any, A>>(model: M): QueryBuilder<A>;
    /**
     * Changes the table this query is currently acting on to the specified table.
     * An optional type parameter can be used to set the scheme of the table so that
     * further operations are as type-safe as possible.
     */
    from<M>(table: string): QueryBuilder<M>;
    /**
     * Creates a new nested where clause. The specified callback receives a nested
     * query builder that can be used to build the nested queries.
     */
    where(handler: (builder: this) => any): this;
    /**
     * Adds a new where clause for the specified column, requiring that it matches
     * the specified variable. This is an alias for where(column, "=", value).
     */
    where<K extends keyof T>(column: K, value: T[K]): this;
    /**
     * Adds a new where clause for the specified column, operator and value. Optionally
     * you can provide a QueryBoolean of "OR" instead of "AND", but it is recommended that
     * you use `orWhere` instead.
     */
    where<K extends keyof T>(column: K, operator: Operator, value: T[K], boolean?: QueryBoolean): this;
    /**
     * Creates a new nested WHERE OR clause. The specified callback receives a nested
     * query builder that can be used to build the nested queries.
     */
    orWhere(handler: (builder: this) => any): this;
    /**
     * Adds a new WHERE OR clause for the specified column, requiring that it matches
     * the specified variable. This is an alias for where(column, "=", value).
     */
    orWhere<K extends keyof T>(column: K, value: T[K]): this;
    /**
     * Adds a new WHERE OR clause for the specified column, operator and value. Use
     * `where` if you want to add a WHERE AND clause instead.
     */
    orWhere<K extends keyof T>(column: K, operator: Operator, value: T[K]): this;
    /**
     * Adds a new where clause comparing two columns of the current query. This uses
     * AND as boolean by default, use `orWhereColumn` if you want OR instead.
     */
    whereColumn<K1 extends keyof T, K2 extends keyof T>(first: K1, operator: Operator, second: K2, boolean?: QueryBoolean): this;
    /**
     * Adds a new WHERE OR clause comparing two columns of the current query. This uses
     * OR, use `whereColumn` if you want AND instead.
     */
    orWhereColumn<K1 extends keyof T, K2 extends keyof T>(first: K1, operator: Operator, second: K2): this;
    /**
     * Adds a new raw where clause for the current query. You can use ? as a substitute for arguments
     * to securely bind parameters.
     */
    whereRaw(query: string, args: DatabaseType[], boolean?: QueryBoolean): this;
    /**
     * Adds a new raw WHERE OR clause for the current query. You can use ? as a substitute for arguments
     * to securely bind parameters.
     */
    orWhereRaw(query: string, args: DatabaseType[]): this;
    /**
     * Adds a new where clause asserting that the specified column is NULL.
     */
    whereNull<K extends keyof T>(column: K, boolean?: QueryBoolean, negate?: boolean): this;
    /**
     * Adds a new WHERE OR clause asserting that the specified column is NULL.
     */
    orWhereNull<K extends keyof T>(column: K): this;
    /**
     * Adds a new where clause asserting that the specified column is NOT NULL.
     */
    whereNotNull<K extends keyof T>(column: K): this;
    /**
     * Adds a new WHERE OR clause asserting that the specified column is NOT NULL.
     */
    orWhereNotNull<K extends keyof T>(column: K): this;
    /**
     * Adds a new nested where query. The handler receives a new QueryBuilder that
     * can be used to enter the where clauses of the nested where.
     */
    whereNested(handler: (builder: QueryBuilder<T>) => void, boolean?: QueryBoolean): this;
    /**
     * Adds a new nested WHERE OR query. The handler receives a new QueryBuilder that
     * can be used to enter the where clauses of the nested where.
     */
    orWhereNested(handler: (builder: QueryBuilder<T>) => void): this;
    /**
     * Adds a new INNER JOIN with the specified table on the specified columns and operators.
     * This operation loses type-safety since it is impossible to determine the return type
     * of the join statically.
     */
    join(table: string, first: string, operator: Operator, second: string, type?: JoinType): QueryBuilder<KeyedDatabaseResult>;
    /**
     * Adds a new nested INNER JOIN with the specified table. The handler receives a JoinClause
     * which it can use to build the new join.
     */
    join(table: string, handler: (clause: JoinClause) => any): QueryBuilder<KeyedDatabaseResult>;
    /**
     * Adds a new INNER JOIN with the specified model and the specified columns and operator.
     * This operation is type safe and returns a union of the current fields and the fields of
     * the specified model.
     */
    join<A, M extends Wrapped<any, A>>(model: M, first: string, operator: Operator, second: string, type?: JoinType): QueryBuilder<T & A>;
    /**
     * Adds a new nested INNER JOIN with the specified model. The handler receives a JoinClause
     * which it can use to build the new join. This operation is type safe and returns a union
     * of the current fields and the fields of the specified model.
     */
    join<A, M extends Wrapped<any, A>>(model: M, handler: (clause: JoinClause) => any): QueryBuilder<T & A>;
    /**
     * Adds a new LEFT JOIN with the specified table on the specified columns and operators.
     * This operation loses type-safety since it is impossible to determine the return type
     * of the join statically.
     */
    leftJoin(table: string, first: string, operator: Operator, second: string): QueryBuilder<KeyedDatabaseResult>;
    /**
     * Adds a new nested LEFT JOIN with the specified table. The handler receives a JoinClause
     * which it can use to build the new join.
     */
    leftJoin(table: string, handler: (clause: JoinClause) => any): QueryBuilder<KeyedDatabaseResult>;
    /**
     * Adds a new LEFT JOIN with the specified model and the specified columns and operator.
     * This operation is type safe and returns a union of the current fields and the fields of
     * the specified model.
     */
    leftJoin<A, M extends Wrapped<any, A>>(model: M, first: string, operator: Operator, second: string): QueryBuilder<T & A>;
    /**
     * Adds a new nested LEFT JOIN with the specified model. The handler receives a JoinClause
     * which it can use to build the new join. This operation is type safe and returns a union
     * of the current fields and the fields of the specified model.
     */
    leftJoin<A, M extends Wrapped<any, A>>(model: M, handler: (clause: JoinClause) => any): QueryBuilder<T & A>;
    /**
     * Adds a new RIGHT JOIN with the specified table on the specified columns and operators.
     * This operation loses type-safety since it is impossible to determine the return type
     * of the join statically.
     */
    rightJoin(table: string, first: string, operator: Operator, second: string): QueryBuilder<KeyedDatabaseResult>;
    /**
     * Adds a new nested RIGHT JOIN with the specified table. The handler receives a JoinClause
     * which it can use to build the new join.
     */
    rightJoin(table: string, handler: (clause: JoinClause) => any): QueryBuilder<KeyedDatabaseResult>;
    /**
     * Adds a new RIGHT JOIN with the specified model and the specified columns and operator.
     * This operation is type safe and returns a union of the current fields and the fields of
     * the specified model.
     */
    rightJoin<A, M extends Wrapped<any, A>>(model: M, first: string, operator: Operator, second: string): QueryBuilder<T & A>;
    /**
     * Adds a new nested RIGHT JOIN with the specified model. The handler receives a JoinClause
     * which it can use to build the new join. This operation is type safe and returns a union
     * of the current fields and the fields of the specified model.
     */
    rightJoin<A, M extends Wrapped<any, A>>(model: M, handler: (clause: JoinClause) => any): QueryBuilder<T & A>;
    /**
     * Marks the query to group the results by the provided column names.
     */
    groupBy<K extends keyof T>(...groups: K[]): this;
    /**
     * Marks the query to order the results by the specified column in the specified
     * direction, or ASC (ascending) by default.
     */
    orderBy<K extends keyof T>(column: K, direction?: OrderDirection): this;
    /**
     * Marks the query to order the results by the specified column in ascending direction.
     */
    orderByAsc<K extends keyof T>(column: K): this;
    /**
     * Marks the query to order the results by the specified column in descending direction.
     */
    orderByDesc<K extends keyof T>(column: K): this;
    /**
     * Limits the amount of rows this query targets to the specified amount. Must be positive non-null.
     */
    limit(count: number): this;
    /**
     * Returns the first row matching this query.
     */
    first(): Promise<T | undefined>;
    /**
     * Finds the first row with the specified ID. Alias for where("id", id).first()
     */
    find(id: number): Promise<T | undefined>;
    /**
     * Returns the specified column of the first result returned by this query.
     */
    value<K extends keyof T>(column: K): Promise<T[K]>;
    /**
     * Alias for get().
     */
    all(): Promise<T[]>;
    /**
     * Returns all rows matching the current query.
     */
    get(): Promise<T[]>;
    /**
     * Returns only the specified column of all rows this query currently targets.
     */
    pluck<K extends keyof T>(column: K): Promise<T[K][]>;
    /**
     * Checks if any rows exist that match the current query. Alias for count > 0.
     */
    exists(): Promise<boolean>;
    /**
     * Counts the amount of rows matching.
     */
    count(): Promise<number>;
    /**
     * Returns the average of the specified column, or the current column by default.
     */
    avg(column?: string): Promise<number>;
    /**
     * Returns the sum of the specified column, or the current column by default.
     */
    sum(column?: string): Promise<number>;
    /**
     * Returns the minimum value of the specified column, or the current column by default.
     */
    min(column?: string): Promise<number>;
    /**
     * Returns the maximum value of the specified column, or the current column by default.
     */
    max(column?: string): Promise<number>;
    /**
     * Runs the specified aggregate function and returns the result.
     */
    aggregate(fn: string, column?: string): Promise<number>;
    /**
     * Inserts the specified entries in the current table.
     */
    insert(...entries: Partial<T>[]): Promise<void>;
    /**
     * Essentially the same as `insert()`, but returns the ID of the inserted
     * row and only supports inserting a single row at a time.
     */
    insertAndGetId(entry: Partial<T>): Promise<number>;
    /**
     * Updates the specified values for all rows matching the current query.
     */
    update(values: Partial<T>): Promise<void>;
    /**
     * Deletes all rows matching the current query.
     */
    delete(): Promise<void>;
    /**
     * Used as a helper function to create a new instance of ourselves. Mainly
     * used in nested where/on clauses. Must be overridden by subclasses.
     */
    protected createNew(): this;
    /**
     * Executes the specified raw query.
     */
    static execute(sql: string, args?: any[]): Promise<void>;
    /**
     * Creates a new QueryBuilder referencing the specified table.
     */
    static table<T extends object>(table: string): QueryBuilder<T>;
    /**
     * Creates a new QueryBuilder referencing the specified model.
     */
    static model<A, T extends Wrapped<any, A>>(model: T): QueryBuilder<A>;
}
import JoinClause from "./join-clause";
