"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const index_1 = require("../index");
/**
 * A fluid SQL query builder that is semi-typesafe and supports selecting, inserting, updating
 * and deleting. Obtain an instance using the static methods or one of the methods on wrapped
 * models.
 */
class QueryBuilder {
    constructor() {
        /**
         * The aggregate function for this query. If this is set, it is assumed
         * that this query selects of format `SELECT fun(columns) AS aggregate`.
         */
        this.aggregateFunction = null;
        /**
         * The columns we are currently retrieving.
         */
        this.columns = ["*"];
        /**
         * If we are only looking for distinct values.
         */
        this.isDistinct = false;
        /**
         * All joins for the current query.
         */
        this.joins = [];
        /**
         * All where clauses for the current query.
         */
        this.wheres = [];
        /**
         * All the GROUP BY columns for the query.
         */
        this.groups = [];
        /**
         * All the ORDER BY clauses for the query.
         */
        this.orders = [];
        /**
         * The LIMIT clause for the query, or -1 if not applicable.
         */
        this.limitCount = -1;
    }
    /**
     * Narrows this query down to the specified fields. This erases type-
     * safety since raw SQL queries can be used. Do not use with untrusted input.
     */
    select(...fields) {
        this.columns = fields;
        this.modelConstructor = null;
        return this;
    }
    /**
     * Marks this query as looking for distinct values. Use distinct(false) to undo.
     */
    distinct(value = true) {
        this.isDistinct = value;
        return this;
    }
    from(modelOrTable) {
        if (typeof modelOrTable === "string") {
            this.table = modelOrTable;
            return this;
        }
        this.table = modelOrTable.tableName;
        this.modelConstructor = modelOrTable;
        this.columns = ["*"];
        return this;
    }
    where(column, operatorOrValue, value, boolean = "AND") {
        if (typeof column === "function") {
            return this.whereNested(column, boolean);
        }
        if (typeof value === "undefined") {
            value = operatorOrValue;
            operatorOrValue = "=";
        }
        this.wheres.push({ type: "basic", column, operator: operatorOrValue, value: value, boolean });
        return this;
    }
    orWhere(column, operatorOrValue, value) {
        return this.where(column, operatorOrValue, value, "OR");
    }
    /**
     * Adds a new where clause comparing two columns of the current query. This uses
     * AND as boolean by default, use `orWhereColumn` if you want OR instead.
     */
    whereColumn(first, operator, second, boolean = "AND") {
        this.wheres.push({
            type: "column",
            first,
            operator,
            second,
            boolean
        });
        return this;
    }
    /**
     * Adds a new WHERE OR clause comparing two columns of the current query. This uses
     * OR, use `whereColumn` if you want AND instead.
     */
    orWhereColumn(first, operator, second) {
        return this.whereColumn(first, operator, second, "OR");
    }
    /**
     * Adds a new raw where clause for the current query. You can use ? as a substitute for arguments
     * to securely bind parameters.
     */
    whereRaw(query, args, boolean = "AND") {
        this.wheres.push({
            type: "raw",
            sql: query,
            values: args,
            boolean
        });
        return this;
    }
    /**
     * Adds a new raw WHERE OR clause for the current query. You can use ? as a substitute for arguments
     * to securely bind parameters.
     */
    orWhereRaw(query, args) {
        return this.whereRaw(query, args, "OR");
    }
    /**
     * Adds a new where clause asserting that the specified column is NULL.
     */
    whereNull(column, boolean = "AND", negate = false) {
        this.wheres.push({
            type: "null",
            column,
            boolean,
            negate
        });
        return this;
    }
    /**
     * Adds a new WHERE OR clause asserting that the specified column is NULL.
     */
    orWhereNull(column) {
        return this.whereNull(column, "OR");
    }
    /**
     * Adds a new where clause asserting that the specified column is NOT NULL.
     */
    whereNotNull(column) {
        return this.whereNull(column, "AND", true);
    }
    /**
     * Adds a new WHERE OR clause asserting that the specified column is NOT NULL.
     */
    orWhereNotNull(column) {
        return this.whereNull(column, "OR", true);
    }
    /**
     * Adds a new nested where query. The handler receives a new QueryBuilder that
     * can be used to enter the where clauses of the nested where.
     */
    whereNested(handler, boolean = "AND") {
        const builder = this.createNew().from(this.table);
        handler(builder);
        this.wheres.push({
            type: "nested",
            builder,
            boolean
        });
        return this;
    }
    /**
     * Adds a new nested WHERE OR query. The handler receives a new QueryBuilder that
     * can be used to enter the where clauses of the nested where.
     */
    orWhereNested(handler) {
        return this.whereNested(handler, "OR");
    }
    join(tableOrModel, first, operator, second, type = "INNER") {
        this.modelConstructor = null;
        const table = typeof tableOrModel === "string" ? tableOrModel : tableOrModel.tableName;
        const clause = new join_clause_1.default(type, table);
        if (typeof first === "function") {
            first(clause);
        }
        else {
            clause.on(first, operator, second);
        }
        this.joins.push(clause);
        return this;
    }
    leftJoin(tableOrModel, first, operator, second) {
        return this.join(tableOrModel, first, operator, second, "LEFT");
    }
    rightJoin(tableOrModel, first, operator, second) {
        return this.join(tableOrModel, first, operator, second, "RIGHT");
    }
    /**
     * Marks the query to group the results by the provided column names.
     */
    groupBy(...groups) {
        this.groups = [...groups, ...this.groups].filter((e, i, a) => a.indexOf(e) === i);
        return this;
    }
    /**
     * Marks the query to order the results by the specified column in the specified
     * direction, or ASC (ascending) by default.
     */
    orderBy(column, direction = "ASC") {
        this.orders.push({ column, direction });
        return this;
    }
    /**
     * Marks the query to order the results by the specified column in ascending direction.
     */
    orderByAsc(column) {
        return this.orderBy(column, "ASC");
    }
    /**
     * Marks the query to order the results by the specified column in descending direction.
     */
    orderByDesc(column) {
        return this.orderBy(column, "DESC");
    }
    /**
     * Limits the amount of rows this query targets to the specified amount. Must be positive non-null.
     */
    limit(count) {
        if (count < 1)
            throw new Error("limit() expects a positive integer.");
        this.limitCount = count;
        return this;
    }
    /**
     * Returns the first row matching this query.
     */
    first() {
        return this.limit(1).get().then(x => x[0]);
    }
    /**
     * Finds the first row with the specified ID. Alias for where("id", id).first()
     */
    find(id) {
        return this.where("id", id).first();
    }
    /**
     * Returns the specified column of the first result returned by this query.
     */
    value(column) {
        return __awaiter(this, void 0, void 0, function* () {
            const entries = yield this.limit(1).get();
            return entries[0][column];
        });
    }
    /**
     * Alias for get().
     */
    all() {
        return this.get();
    }
    /**
     * Returns all rows matching the current query.
     */
    get() {
        const compiler = index_1.default.getEngine().getGrammarCompiler();
        const query = compiler.compileSelect(this);
        return index_1.default.getEngine().get(query.sql, query.args).then(rows => {
            return rows.map(row => {
                if (this.modelConstructor) {
                    const instance = new this.modelConstructor();
                    return Object.assign(instance, row);
                }
                return row;
            });
        }).then(Object.freeze);
    }
    /**
     * Returns only the specified column of all rows this query currently targets.
     */
    pluck(column) {
        this.columns = [column];
        return this.get().then(x => x.map(y => y[column]));
    }
    /**
     * Checks if any rows exist that match the current query. Alias for count > 0.
     */
    exists() {
        return this.count().then(count => count > 0);
    }
    /**
     * Counts the amount of rows matching.
     */
    count() {
        return this.aggregate("COUNT");
    }
    /**
     * Returns the average of the specified column, or the current column by default.
     */
    avg(column) {
        return this.aggregate("AVG", column);
    }
    /**
     * Returns the sum of the specified column, or the current column by default.
     */
    sum(column) {
        return this.aggregate("SUM", column);
    }
    /**
     * Returns the minimum value of the specified column, or the current column by default.
     */
    min(column) {
        return this.aggregate("MIN", column);
    }
    /**
     * Returns the maximum value of the specified column, or the current column by default.
     */
    max(column) {
        return this.aggregate("MAX", column);
    }
    /**
     * Runs the specified aggregate function and returns the result.
     */
    aggregate(fn, column) {
        this.columns = column ? [column] : this.columns;
        this.aggregateFunction = fn;
        const compiler = index_1.default.getEngine().getGrammarCompiler();
        const query = compiler.compileSelect(this);
        return index_1.default.getEngine().get(query.sql, query.args).then(x => x[0]["aggregate"]);
    }
    /**
     * Inserts the specified entries in the current table.
     */
    insert(...entries) {
        const compiler = index_1.default.getEngine().getGrammarCompiler();
        const query = compiler.compileInsert(this, entries);
        return index_1.default.getEngine().query(query.sql, query.args);
    }
    /**
     * Essentially the same as `insert()`, but returns the ID of the inserted
     * row and only supports inserting a single row at a time.
     */
    insertAndGetId(entry) {
        const compiler = index_1.default.getEngine().getGrammarCompiler();
        const query = compiler.compileInsert(this, [entry]);
        return index_1.default.getEngine().insertAndGetId(this.table, query.sql, query.args);
    }
    /**
     * Updates the specified values for all rows matching the current query.
     */
    update(values) {
        const compiler = index_1.default.getEngine().getGrammarCompiler();
        const query = compiler.compileUpdate(this, values);
        return index_1.default.getEngine().query(query.sql, query.args);
    }
    /**
     * Deletes all rows matching the current query.
     */
    delete() {
        const compiler = index_1.default.getEngine().getGrammarCompiler();
        const query = compiler.compileDelete(this);
        return index_1.default.getEngine().query(query.sql, query.args);
    }
    /**
     * Used as a helper function to create a new instance of ourselves. Mainly
     * used in nested where/on clauses. Must be overridden by subclasses.
     */
    createNew() {
        return new QueryBuilder();
    }
    /**
     * Executes the specified raw query.
     */
    static execute(sql, args) {
        return index_1.default.getEngine().query(sql, args || []);
    }
    /**
     * Creates a new QueryBuilder referencing the specified table.
     */
    static table(table) {
        return new QueryBuilder().from(table);
    }
    /**
     * Creates a new QueryBuilder referencing the specified model.
     */
    static model(model) {
        return new QueryBuilder().from(model);
    }
}
exports.default = QueryBuilder;
// This needs to be here (below QueryBuilder) to prevent a cyclic dependency.
const join_clause_1 = require("./join-clause");
