import QueryBuilder from "./builder";
import { ColumnWhereClause, NullWhereClause, RawWhereClause, SimpleWhereClause, NestedWhereClause } from "./types";
import { DatabaseType } from "../base-model";
export declare type Builder = QueryBuilder<any>;
export declare type QueryComponent = {
    sql: string;
    args: DatabaseType[];
};
/**
 * This class is responsible for "compiling" QueryBuilders into the raw SQL
 * that will be executed by the database engine. It returns both SQL and bound
 * values, so we can safely let the database engine handle escaping of arguments.
 */
export default class SQLGrammarCompiler {
    /**
     * Compiles a full SELECT query.
     */
    compileSelect(builder: Builder): QueryComponent;
    /**
     * Compiles all components of a normal SELECT query. This can be overridden by subclasses
     * to reorder this or to add/remove components.
     */
    protected compileComponents(builder: Builder): {
        aggregate: QueryComponent;
        columns: QueryComponent;
        from: QueryComponent;
        joins: QueryComponent;
        wheres: QueryComponent;
        groups: QueryComponent;
        orders: QueryComponent;
        limit: QueryComponent;
    };
    /**
     * Compiles the aggregate function part of a SELECT query.
     */
    compileAggregate(builder: Builder): QueryComponent;
    /**
     * Compiles the SELECT `columns` part of a SELECT query, unless
     * an aggregate function was specified.
     */
    compileColumns(builder: Builder): QueryComponent;
    /**
     * Compiles the FROM `table` part of a SELECT query.
     */
    compileFrom(builder: Builder): QueryComponent;
    /**
     * Compiles all JOIN conditions of the query.
     */
    compileJoins(builder: Builder): QueryComponent;
    /**
     * Compiles all WHERE clauses of the query.
     */
    compileWheres(builder: Builder): QueryComponent;
    /**
     * Compiles the set of WHERE conditions in the specified query. Used
     * for both JOIN and WHERE compiling.
     */
    compileConditions(builder: Builder): QueryComponent;
    /**
     * Compiles a simple WHERE clause.
     */
    compileBasicWhere(builder: Builder, where: SimpleWhereClause): QueryComponent;
    /**
     * Compiles a WHERE clause comparing two columns.
     */
    compileColumnWhere(builder: Builder, where: ColumnWhereClause): QueryComponent;
    /**
     * Compiles a raw SQL where clause.
     */
    compileRawWhere(builder: Builder, where: RawWhereClause): QueryComponent;
    /**
     * Compiles a WHERE clause checking for (NOT) NULL.
     */
    compileNullWhere(builder: Builder, where: NullWhereClause): QueryComponent;
    /**
     * Compiles a nested WHERE clause.
     */
    compileNestedWhere(builder: Builder, where: NestedWhereClause): QueryComponent;
    /**
     * Compiles the GROUP BY clauses of the query.
     */
    compileGroups(builder: Builder): QueryComponent;
    /**
     * Compile the ORDER BY clauses of the query.
     */
    compileOrders(builder: Builder): QueryComponent;
    /**
     * Compiles the LIMIT part of the query.
     */
    compileLimit(builder: Builder): QueryComponent;
    /**
     * Compiles a full INSERT query for the specified values. This accepts either a model
     * or a keyed database value object. This assumes that every element in the `values`
     * array has the same structure, and that it is not an empty array.
     */
    compileInsert<T>(builder: QueryBuilder<T>, values: T[]): QueryComponent;
    /**
     * Compiles a full UPDATE query for the specified partial. This does not diff-check and
     * simply assumes that everything specified needs to be updated.
     */
    compileUpdate<T>(builder: QueryBuilder<T>, value: Partial<T>): QueryComponent;
    /**
     * Compiles a full DELETE query for all rows matching the builder's where clauses.
     */
    compileDelete(builder: Builder): QueryComponent;
    /**
     * Escapes the specified column name. This is intended to be overridden by potential subclasses.
     */
    protected escapeColumn(column: string): string;
    /**
     * Escapes the specified table name. This is intended to be overridden by potential subclasses.
     */
    protected escapeTable(table: string): string;
}
