"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
/**
 * This class is responsible for "compiling" QueryBuilders into the raw SQL
 * that will be executed by the database engine. It returns both SQL and bound
 * values, so we can safely let the database engine handle escaping of arguments.
 */
class SQLGrammarCompiler {
    /**
     * Compiles a full SELECT query.
     */
    compileSelect(builder) {
        const components = this.compileComponents(builder);
        return {
            sql: Object.keys(components).map(x => components[x].sql).filter(x => !!x).join(" "),
            args: Object.keys(components).map(x => components[x].args).reduce((p, c) => [...p, ...c], [])
        };
    }
    /**
     * Compiles all components of a normal SELECT query. This can be overridden by subclasses
     * to reorder this or to add/remove components.
     */
    compileComponents(builder) {
        return {
            aggregate: this.compileAggregate(builder),
            columns: this.compileColumns(builder),
            from: this.compileFrom(builder),
            joins: this.compileJoins(builder),
            wheres: this.compileWheres(builder),
            groups: this.compileGroups(builder),
            orders: this.compileOrders(builder),
            limit: this.compileLimit(builder)
        };
    }
    /**
     * Compiles the aggregate function part of a SELECT query.
     */
    compileAggregate(builder) {
        if (!builder.aggregateFunction)
            return { sql: "", args: [] };
        const columns = builder.columns.map(x => this.escapeColumn(x)).join(", ");
        return {
            sql: `SELECT ${builder.aggregateFunction}(${builder.isDistinct && columns !== "*" ? "DISTINCT " : ""}${columns}) AS aggregate`,
            args: []
        };
    }
    /**
     * Compiles the SELECT `columns` part of a SELECT query, unless
     * an aggregate function was specified.
     */
    compileColumns(builder) {
        // We will let aggregate start the query if its present.
        if (builder.aggregateFunction)
            return { sql: "", args: [] };
        return {
            sql: (builder.isDistinct ? "SELECT DISTINCT " : "SELECT ") + builder.columns.map(c => this.escapeColumn(c)).join(", "),
            args: []
        };
    }
    /**
     * Compiles the FROM `table` part of a SELECT query.
     */
    compileFrom(builder) {
        return {
            sql: "FROM " + this.escapeTable(builder.table),
            args: []
        };
    }
    /**
     * Compiles all JOIN conditions of the query.
     */
    compileJoins(builder) {
        const parts = builder.joins.map(join => {
            const condition = this.compileConditions(join);
            return {
                sql: `${join.type} JOIN ${this.escapeTable(join.joiningOn)} ON ${condition.sql}`,
                args: condition.args
            };
        });
        return {
            sql: parts.map(x => x.sql).join(" "),
            args: parts.map(x => x.args).reduce((p, c) => [...p, ...c], [])
        };
    }
    /**
     * Compiles all WHERE clauses of the query.
     */
    compileWheres(builder) {
        if (!builder.wheres.length)
            return { sql: "", args: [] };
        const part = this.compileConditions(builder);
        return {
            sql: "WHERE " + part.sql,
            args: part.args
        };
    }
    /**
     * Compiles the set of WHERE conditions in the specified query. Used
     * for both JOIN and WHERE compiling.
     */
    compileConditions(builder) {
        /* istanbul ignore next This should be caught by compileWheres, but is here as a sanity check. */
        if (!builder.wheres.length)
            return { sql: "", args: [] };
        const parts = builder.wheres.map((where, i) => {
            let content;
            if (where.type === "basic")
                content = this.compileBasicWhere(builder, where);
            else if (where.type === "column")
                content = this.compileColumnWhere(builder, where);
            else if (where.type === "raw")
                content = this.compileRawWhere(builder, where);
            else if (where.type === "null")
                content = this.compileNullWhere(builder, where);
            else if (where.type === "nested")
                content = this.compileNestedWhere(builder, where);
            /* istanbul ignore next Impossible with a properly formed querybuilder. */
            if (!content) {
                throw new Error("Invalid type for where clause.");
            }
            return {
                sql: (i === 0 ? "" : where.boolean + " ") + content.sql,
                args: content.args
            };
        });
        return {
            sql: parts.map(x => x.sql).join(" "),
            args: parts.map(x => x.args).reduce((p, c) => [...p, ...c], [])
        };
    }
    /**
     * Compiles a simple WHERE clause.
     */
    compileBasicWhere(builder, where) {
        return {
            sql: `${this.escapeColumn(where.column)} ${where.operator} ?`,
            args: [where.value]
        };
    }
    /**
     * Compiles a WHERE clause comparing two columns.
     */
    compileColumnWhere(builder, where) {
        return {
            sql: `${this.escapeColumn(where.first)} ${where.operator} ${this.escapeColumn(where.second)}`,
            args: []
        };
    }
    /**
     * Compiles a raw SQL where clause.
     */
    compileRawWhere(builder, where) {
        return {
            sql: where.sql,
            args: where.values
        };
    }
    /**
     * Compiles a WHERE clause checking for (NOT) NULL.
     */
    compileNullWhere(builder, where) {
        return {
            sql: this.escapeColumn(where.column) + " IS " + (where.negate ? "NOT " : "") + "NULL",
            args: []
        };
    }
    /**
     * Compiles a nested WHERE clause.
     */
    compileNestedWhere(builder, where) {
        const ret = this.compileConditions(where.builder);
        return {
            sql: "(" + ret.sql + ")",
            args: ret.args
        };
    }
    /**
     * Compiles the GROUP BY clauses of the query.
     */
    compileGroups(builder) {
        if (!builder.groups.length)
            return { sql: "", args: [] };
        return {
            sql: "GROUP BY " + builder.groups.map(x => this.escapeColumn(x)).join(", "),
            args: []
        };
    }
    /**
     * Compile the ORDER BY clauses of the query.
     */
    compileOrders(builder) {
        if (!builder.orders.length)
            return { sql: "", args: [] };
        return {
            sql: "ORDER BY " + builder.orders.map(x => this.escapeColumn(x.column) + " " + x.direction).join(", "),
            args: []
        };
    }
    /**
     * Compiles the LIMIT part of the query.
     */
    compileLimit(builder) {
        if (builder.limitCount === -1)
            return { sql: "", args: [] };
        return {
            sql: "LIMIT ?",
            args: [builder.limitCount]
        };
    }
    /**
     * Compiles a full INSERT query for the specified values. This accepts either a model
     * or a keyed database value object. This assumes that every element in the `values`
     * array has the same structure, and that it is not an empty array.
     */
    compileInsert(builder, values) {
        const keys = Object.keys(values[0]).filter(x => x !== "id" && typeof values[0][x] !== "function");
        const columns = keys.map(x => this.escapeColumn(x)).join(",");
        const args = values.map(x => keys.map(k => x[k])).reduce((p, c) => [...p, ...c], []);
        const placeholders = values.map(x => "(" + keys.map(k => "?").join(",") + ")").join(", ");
        return {
            sql: `INSERT INTO ${this.escapeTable(builder.table)} (${columns}) VALUES ${placeholders}`,
            args
        };
    }
    /**
     * Compiles a full UPDATE query for the specified partial. This does not diff-check and
     * simply assumes that everything specified needs to be updated.
     */
    compileUpdate(builder, value) {
        const columns = Object.keys(value).filter(x => x !== "id" && typeof value[x] !== "function").map(x => x + " = ?").join(", ");
        const args = Object.keys(value).filter(x => x !== "id" && typeof value[x] !== "function").map(x => value[x]);
        const joins = this.compileJoins(builder);
        const wheres = this.compileWheres(builder);
        return {
            sql: `UPDATE ${this.escapeTable(builder.table)}${joins.sql ? " " + joins.sql : ""} SET ${columns} ${wheres.sql}`,
            args: [...joins.args, ...args, ...wheres.args]
        };
    }
    /**
     * Compiles a full DELETE query for all rows matching the builder's where clauses.
     */
    compileDelete(builder) {
        const wheres = this.compileWheres(builder);
        return {
            sql: `DELETE FROM ${this.escapeTable(builder.table)}${wheres.sql ? " " + wheres.sql : ""}`,
            args: wheres.args
        };
    }
    /**
     * Escapes the specified column name. This is intended to be overridden by potential subclasses.
     */
    escapeColumn(column) {
        return column;
    }
    /**
     * Escapes the specified table name. This is intended to be overridden by potential subclasses.
     */
    escapeTable(table) {
        return table;
    }
}
exports.default = SQLGrammarCompiler;
